from typing import Dict, List, Union, Optional, Callable, Any
from scabha.cargo import Parameter, _UNSET_DEFAULT, EmptyListDefault
from scabha.cargo import EmptyDictDefault, ParameterPolicies, Cargo
from scabha.basetypes import  File, Directory
from scabha.validate import validate_parameters
from dataclasses import dataclass
from omegaconf import OmegaConf, ListConfig
from pydoc import locate


@dataclass
class OldParameter:
    name: str 
    dtype: Any
    info: str
    default: Any = None
    required: bool = False
    choices: Optional[List[Any]] = None
    io: Optional[str] = ""
    mapping: Optional[str] = ""
    check_io: bool = False
    deprecated: bool = False
    positional: bool = False


@dataclass
class OldCab:
    task: str
    base: str
    version: List[str]
    binary: str = ""
    description: str = "<documentation>"
    prefix: str = "--"
    parameters: Optional[List[Dict]] = None
    tag: Optional[List[str]] = None
    junk: Optional[List[str]] = None
    msdir: bool = False
    wranglers: Optional[List[str]] = None

class SimpleCab:
    def __init__(self, oldfile: File):
        self.oldfile = oldfile
        cab_strct = OmegaConf.structured(OldCab)
        param_strct = OmegaConf.structured(OldParameter)
        _oldcab = OmegaConf.load(oldfile)
        self.oldcab = OmegaConf.merge(cab_strct,
                                    OmegaConf.load(oldfile))
        self.parameters = []
        for param in _oldcab.parameters:
            pardict = OmegaConf.merge(param_strct, param)
            self.parameters.append(pardict)
            
        
    
    def __to_new_dtype(self, param:OldParameter) -> str:
        new_dtype = []
        if isinstance(param.dtype, (list, ListConfig)):
            islist = True
            dtype = param.dtype
        else:
            islist = False
            dtype = [param.dtype]
            
        for item in dtype:
            listof = False
            if item.startswith("list:"):
                item = item.split(":")[-1] 
                listof = True
    
            if item == "file":
                if param.get("io", "") == "msfile":
                    newtype = "MS"
                else:
                    newtype = "File"
            else:
                newtype = item
            
            if listof:
                new_dtype.append(f"List[{newtype}]")
            else:
                new_dtype.append(newtype)
        if islist:
            return "Union[" + ",".join(new_dtype) + "]"
        else:
            return new_dtype[0]
        
    
    def to_new_params(self, set_inputs=True):
        """AI is creating summary for init_from_old_cab

        Args:
            oldcab (File): [description]
        """
        param_struct = OmegaConf.structured(Parameter)
        params = {}
        for param in self.parameters:
            dtype = self.__to_new_dtype(param)
            
            policies = ParameterPolicies(positional=param.positional)
            
            params[param.name] = OmegaConf.merge(param_struct,
                            dict(info=param.info, dtype=dtype, policies=policies,
                                nom_de_guerre=param.mapping,
                                must_exist=param.check_io))
            if set_inputs:
                self.inputs = params
                self.outputs = {}
        return OmegaConf.create(params)
    
    
    def save(self, path: str):
        """_summary_

        Args:
            path (str): _description_

        Returns:
            _type_: _description_
        """
        
        outdict = OmegaConf.create({"inputs": self.inputs,
                                    "outputs": self.outputs,
                                    })
        OmegaConf.save(outdict, path)
        
        return 0

        
        
        
        
        
    
    

    


import os
from typing import Any
from omegaconf import OmegaConf, DictConfig


ENVVAR_SYNTAX = "ENV::"

def substitute_environment_variables(element: Any):
    """Find and replace occurences of environment varaibles in dict values."""
    if isinstance(element, (dict, DictConfig)):
        for k, v in element.items():
            element[k] = substitute_environment_variables(v)
    elif isinstance(element, str) and element.startswith(ENVVAR_SYNTAX):
        envvar = element.lstrip(ENVVAR_SYNTAX)
        if envvar not in os.environ:
            raise KeyError(f"Environment variable {envvar} is not set.")
        return os.environ[envvar]
    return element

def resolve_version_substitutions(config):
    """Resolve version substitutions using metadata and assign sections."""
    
    for image_name, image in config.images.items():

        lookup_dict = OmegaConf.merge(dict(**config.metadata), config.assign)
        lookup_dict.update(**(image.assign or {}))
    
        resolved_versions = {}  # Recreate to preserve ordering.
        for version_name, version in image.versions.items():
            try:
                resolved_name = version_name.format(**lookup_dict)
            except KeyError as e:
                msg = (
                    f"Unable to resolve substitution '{version_name}' in "
                    f"versions field of manifest for image '{image_name}'."
                )
                raise Exception(msg) from e
            resolved_versions[resolved_name] = version

        image.versions = resolved_versions

    return




import os.path
from typing import Dict, Any
from omegaconf import OmegaConf
from dataclasses import dataclass

from scabha.cargo import Parameter

@dataclass
class JonesTemplateStructure:
    JONES_TEMPLATE: Dict[str, Parameter]

JonesTemplate = None

if JonesTemplate is None:
    _dirname = os.path.dirname(__file__)
    _structure = OmegaConf.create(JonesTemplateStructure)
    _config = OmegaConf.load(os.path.join(
        os.path.dirname(__file__),
        "schema_JONES_TEMPLATE.yaml"))
    JonesTemplate = OmegaConf.merge(_structure, _config).JONES_TEMPLATE

def make_stimela_schema(params: Dict[str, Any], inputs: Dict[str, Parameter], outputs: Dict[str, Parameter]):
    """Augments a schema for stimela based on solver.terms"""
    inputs = inputs.copy()

    terms = params.get('sol.jones', [])

    for jones in terms:
        for key, value in JonesTemplate.items():
            inputs[f"{jones.lower()}.{key}"] = value
        # inputs[f"{jones}.label"].default = jones

    return inputs, outputs



# needs pip install "ruamel.yaml<0.18.0" for newer pythoin versions
import requests
from ruamel.yaml import YAML
yaml = YAML()
yaml.indent(mapping=2, sequence=4, offset=2)
import cultcargo

mpath = cultcargo.__file__.rstrip('__init__.py') + 'genesis/pfb-imaging/latest'

branch = 'main'
content_path = f'https://raw.githubusercontent.com/ratt-ru/pfb-imaging/{branch}/pfb/parser'

# returns list of files in subfolder
response = requests.get(f'https://api.github.com/repos/ratt-ru/pfb-imaging/contents/pfb/parser?ref={branch}')
for r in response.json():
    name = r['name']
    print(name)
    if name.endswith('yml') or name.endswith('yaml'):
        # get file content
        config = requests.get(f'{content_path}/{name}')
        # round trip to get correct formatting
        data = yaml.load(config.text)
        # dump to yaml
        with open(f'{mpath}/{name}', 'w') as f:
            yaml.dump(data, f)

print('Done')



from dataclasses import fields, _MISSING_TYPE
from typing import List, Dict
from .converters import as_time, as_freq


class BaseConfigSection:
    """Base class for dynamically generated config dataclasses."""

    def __validate_choices__(self):

        for fld in fields(self):
            name = fld.name
            value = getattr(self, name)
            meta = fld.metadata

            if value is None or "choices" not in meta:
                continue
            else:
                choices = meta.get("choices")
                assert value in choices, (
                    f"Invalid input in {name}. User specified '{value}'. "
                    f"Valid choices are {choices}."
                )

    def __validate_element_choices__(self):

        for fld in fields(self):
            name = fld.name
            value = getattr(self, name)
            meta = fld.metadata

            if value is None or "element_choices" not in meta:
                continue
            else:
                element_choices = meta.get("element_choices")
                if isinstance(value, List):
                    elements = value
                elif isinstance(value, Dict):
                    elements = value.values()
                else:
                    raise ValueError(
                        f"Parameter {name} of type {type(value)} has element "
                        f"choices. This is not supported."
                    )
                invalid_elements = set(elements) - set(element_choices)
                assert not invalid_elements, (
                    f"Invalid input in {name}. User specified "
                    f"{elements}. Valid choices: {element_choices}."
                )

    def __helpstr__(self):

        helpstrings = {}

        for fld in fields(self):

            meta = fld.metadata

            help_str = meta.get('help')

            if meta.get("choices", None):
                choice_str = f"Choices: {meta.get('choices')} "
            else:
                choice_str = ""

            if meta.get("element_choices", None):
                element_choice_str = f"Choices: {meta.get('element_choices')} "
            else:
                element_choice_str = ""

            if isinstance(fld.default, _MISSING_TYPE):
                default = fld.default_factory()
            else:
                default = fld.default

            if fld.metadata.get('required', False):
                default_str = "REQUIRED"
            else:
                default_str = f"Default: {default}"

            helpstrings[fld.name] = (
                f"{help_str} {choice_str}{element_choice_str}{default_str}"
            )

        return helpstrings


def __input_ms_post_init__(self):

    self.__validate_choices__()
    self.__validate_element_choices__()
    self.time_chunk = as_time(self.time_chunk)
    self.freq_chunk = as_freq(self.freq_chunk)

    assert len(self.select_uv_range) == 2, \
        "input_ms.select_uv_range expects a two-element list."

    assert not (self.sigma_column and self.weight_column), \
        "sigma_column and weight_column are mutually exclusive."

    if self.is_bda:
        assert self.time_chunk == 0, \
            ("input_ms.is_bda does not support chunking in time. Please "
                "set input_ms.time_chunk to 0.")
        assert self.freq_chunk == 0, \
            ("input_ms.is_bda does not support chunking in freq. Please "
                "set input_ms.freq_chunk to 0.")


def __input_model_post_init__(self):
    self.__validate_choices__()
    self.__validate_element_choices__()


def __output_post_init__(self):

    self.__validate_choices__()
    self.__validate_element_choices__()
    assert not (bool(self.products) ^ bool(self.columns)), \
        "Neither or both of products and columns must be specified."
    if self.products:
        assert len(self.products) == len(self.columns), \
                "Number of products not equal to number of columns."

    if self.net_gains:
        nested = any(isinstance(i, list) for i in self.net_gains)
        if nested:
            assert all(isinstance(i, list) for i in self.net_gains), \
                ("Contents of outputs.net_gains not understood. "
                    "Must be strictly a list or list of lists.")
        else:
            assert all(isinstance(i, str) for i in self.net_gains), \
                ("Contents of outputs.net_gains not understood. "
                    "Must be strictly a list or list of lists.")
            # In the non-nested case, introduce outer list (consistent).
            self.net_gains = [self.net_gains]

    if isinstance(self.subtract_directions, list):
        self.subtract_directions = tuple(self.subtract_directions)


def __mad_flags_post_init__(self):
    self.__validate_choices__()
    self.__validate_element_choices__()


def __solver_post_init__(self):
    self.__validate_choices__()
    self.__validate_element_choices__()
    assert len(self.iter_recipe) >= len(self.terms), \
        "User has specified solver.iter_recipe with too few elements."

    assert self.convergence_criteria >= 1e-8, \
        "User has specified solver.convergence_criteria below 1e-8."


def __dask_post_init__(self):
    self.__validate_choices__()
    self.__validate_element_choices__()
    if self.address:
        msg = (
            "Scheduler address supplied but dask.scheduler has not "
            "been set to distributed."
        )
        assert self.scheduler == "distributed", msg


def __gain_post_init__(self):
    self.__validate_choices__()
    self.__validate_element_choices__()
    self.time_interval = as_time(self.time_interval)
    self.freq_interval = as_freq(self.freq_interval)
    if self.type == "crosshand_phase" and self.solve_per != "array":
        raise ValueError(
            "Crosshand phase can only be solved as a per array term. Please "
            "set the appropriate term.solve_per to 'array'."
        )
    if isinstance(self.pinned_directions, list):
        self.pinned_directions = tuple(self.pinned_directions)


POST_INIT_MAP = {
    "input_ms": __input_ms_post_init__,
    "input_model": __input_model_post_init__,
    "output": __output_post_init__,
    "mad_flags": __mad_flags_post_init__,
    "solver": __solver_post_init__,
    "dask": __dask_post_init__,
    "gain": __gain_post_init__
}
